from jupyter_server.base.handlers import APIHandler
from tornado import web
import json
from .utils.open_ai_utils import get_open_ai_completion


# This handler is responsible for the mito_ai/completion endpoint. 
# It takes a message from the user, sends it to the OpenAI API, and returns the response.
# Important: Because this is a server extension, print statements are sent to the 
# jupyter server terminal by default (ie: the terminal you ran `jupyter lab`)
class OpenAICompletionHandler(APIHandler):
    @web.authenticated
    def post(self):
        # Retrieve the message from the request
        data = self.get_json_body()
        messages = data.get('messages', '')

        try:
            # Query OpenAI API
            response = get_open_ai_completion(messages)
            self.finish(json.dumps(response))
        except PermissionError as e:
            # Raise a PermissionError when the user has 
            # reached the free tier limit for Mito AI.
            self.set_status(403)
            self.finish()
        except Exception as e:
            # Catch all other exceptions and return a 500 error
            self.set_status(500)
            self.finish()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.

import os
import requests
from typing import Any, Dict, List
from .db import get_user_field, set_user_field
from .schema import UJ_AI_MITO_API_NUM_USAGES, UJ_STATIC_USER_ID, UJ_USER_EMAIL
from .version_utils import is_pro
from .create import initialize_user
from .telemetry_utils import log, KEY_TYPE_PARAM, MITO_SERVER_KEY, USER_KEY, MITO_AI_COMPLETION_SUCCESS, MITO_AI_COMPLETION_ERROR, MITO_SERVER_NUM_USAGES, MITO_SERVER_FREE_TIER_LIMIT_REACHED
OPEN_AI_URL = 'https://api.openai.com/v1/chat/completions'
MITO_AI_URL = 'https://ogtzairktg.execute-api.us-east-1.amazonaws.com/Prod/completions/'

OPEN_SOURCE_AI_COMPLETIONS_LIMIT = 500

__user_email = None
__user_id = None
__num_usages = None

def _get_ai_completion_data(messages: List[Dict[str, Any]]) -> Dict[str, Any]:
        return {
            "model": "gpt-4o-mini",
            "messages": messages,
            "temperature": 0,
        }

def _get_ai_completion_with_key(ai_completion_data: Dict[str, Any], OPENAI_API_KEY: str) -> Dict[str, Any]:
        headers = {
                'Content-Type': 'application/json',
                'Authorization': f'Bearer {OPENAI_API_KEY}' 
        }

        res = requests.post(OPEN_AI_URL, headers=headers, json=ai_completion_data)

        # If the response status code is in the 200s, this does nothing
        # If the response status code indicates an error (4xx or 5xx), 
        # raise an HTTPError exception with details about what went wrong
        res.raise_for_status()

        completion = res.json()['choices'][0]['message']['content']
        return {'completion': completion}
        

def _get_ai_completion_from_mito_server(last_message_content: str, ai_completion_data: Dict[str, Any]) -> Dict[str, Any]:
        
        global __user_email, __user_id, __num_usages

        if __user_email is None:
                __user_email = get_user_field(UJ_USER_EMAIL)
        if __user_id is None:
                __user_id = get_user_field(UJ_STATIC_USER_ID)
        if __num_usages is None:
                __num_usages = get_user_field(UJ_AI_MITO_API_NUM_USAGES)

        if __num_usages is None:
                __num_usages = 0

        pro = is_pro()

        if not pro and __num_usages >= OPEN_SOURCE_AI_COMPLETIONS_LIMIT:
                log(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
                raise PermissionError(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
                
        data = {
                'email': __user_email,
                'user_id': __user_id,
                'data': ai_completion_data,
                'user_input': last_message_content # We add this just for logging purposes
        }

        headers = {
                'Content-Type': 'application/json',
        }

        res = requests.post(MITO_AI_URL, headers=headers, json=data)

        # If the response status code is in the 200s, this does nothing
        # If the response status code indicates an error (4xx or 5xx), 
        # raise an HTTPError exception with details about what went wrong
        res.raise_for_status()

        # The lambda function returns a dictionary with a completion entry in it,
        # so we just return that.
        return res.json()

        

def get_open_ai_completion(messages: List[Dict[str, Any]]) -> Dict[str, Any]:

        initialize_user()

        OPENAI_API_KEY = os.environ.get('OPENAI_API_KEY')
                
        # Prep the ai completion data 
        ai_completion_data = _get_ai_completion_data(messages)
        last_message_content = messages[-1]['content']

        # Try to get the AI response
        try:
                if OPENAI_API_KEY is None:
                        # If they don't have an Open AI key, use the mito server to get a completion
                        response = _get_ai_completion_from_mito_server(last_message_content, ai_completion_data)

                        # Increment the number of usages
                        global __num_usages
                        __num_usages = __num_usages + 1
                        set_user_field(UJ_AI_MITO_API_NUM_USAGES, __num_usages)

                        # Log the successful completion
                        log(MITO_AI_COMPLETION_SUCCESS, params={
                                KEY_TYPE_PARAM: MITO_SERVER_KEY,
                                MITO_SERVER_NUM_USAGES: __num_usages
                        })
                        return response
                else:
                        # If they DO have an Open AI key, use it to get a completion
                        response = _get_ai_completion_with_key(ai_completion_data, OPENAI_API_KEY)

                        # Log the successful completion
                        log(MITO_AI_COMPLETION_SUCCESS, params={KEY_TYPE_PARAM: USER_KEY})
                        return response
        except Exception as e:
                key_type = MITO_SERVER_KEY if OPENAI_API_KEY is None else USER_KEY
                log(
                        MITO_AI_COMPLETION_ERROR, 
                        params={
                                KEY_TYPE_PARAM: key_type
                        }, 
                        error=e
                )
                raise e


from jupyter_server.utils import url_path_join
from .OpenAICompletionHandler import OpenAICompletionHandler
from pathlib import Path
import json

try:
    from _version import __version__
except ImportError:
    # Fallback when using the package in dev mode without installing
    # in editable mode with pip. It is highly recommended to install
    # the package from a stable release or in editable mode: https://pip.pypa.io/en/stable/topics/local-project-installs/#editable-installs
    import warnings
    warnings.warn("Importing 'mito-ai' outside a proper installation.")
    __version__ = "dev"


HERE = Path(__file__).parent.resolve()


with (HERE / "labextension" / "package.json").open() as fid:
    data = json.load(fid)

def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": "mito-ai"
    }]

def _jupyter_server_extension_points():
    """
    Returns a list of dictionaries with metadata describing
    where to find the `_load_jupyter_server_extension` function.
    """
    return [{"module": "mito-ai"}]

# Jupyter Server is the backend used by JupyterLab. A sever extension lets 
# us add new API's to the backend, so we can do some processing that we don't
# want to exist in the users's javascript. 

# For a further explanation of the Jupyter architecture watch the first 35 minutes
# of this video: https://www.youtube.com/watch?v=9_-siU-_XoI
def _load_jupyter_server_extension(server_app):
    host_pattern = ".*$"
    web_app = server_app.web_app
    base_url = web_app.settings["base_url"]
    route_pattern = url_path_join(base_url, "mito_ai", "completion")
    handlers = [(route_pattern, OpenAICompletionHandler)]
    web_app.add_handlers(host_pattern, handlers)
    web_app.log.info("Loaded the mito_ai server extension")



import os
from typing import Any, Dict, Optional
import requests
from .version_utils import MITOSHEET_HELPER_PRIVATE, is_pro
from .schema import UJ_MITOSHEET_TELEMETRY, UJ_STATIC_USER_ID, UJ_USER_EMAIL, UJ_FEEDBACKS_V2
from .db import get_user_field
from .._version import __version__
from .utils import is_running_test

import analytics
WRITE_KEY = '6I7ptc5wcIGC4WZ0N1t0NXvvAbjRGUgX' 
analytics.write_key = WRITE_KEY

# If you want, you can optionally choose to print logs
# helpful for debugging.
PRINT_LOGS = False

################################# 
# Mito AI Completion 
# Constants for logging the success or error of Mito AI
MITO_AI_COMPLETION_SUCCESS = 'mito_ai_success'
MITO_AI_COMPLETION_ERROR = 'mito_ai_error'

# Params 
# - logging the type of key 
KEY_TYPE_PARAM = 'AI_key_type'
MITO_SERVER_KEY= 'mito_server_key'
USER_KEY = 'user_key'

# - logging the number of usages of the Mito server
MITO_SERVER_NUM_USAGES = 'mito_server_num_usages'
#################################

################################# 
# Mito Server Free Tier Reached
MITO_SERVER_FREE_TIER_LIMIT_REACHED = 'mito_server_free_tier_limit_reached'
#################################

def telemetry_turned_on() -> bool:
    """
    Helper function that tells you if logging is turned on or
    turned off on the entire Mito instance
    """
    # If private helper is installed, then we don't log anything
    if MITOSHEET_HELPER_PRIVATE:
        return False

    # TODO: Check if the an enterprise user has turned telemetry to true

    # If Mito Pro is on, then don't log anything
    if is_pro():
        return False

    telemetry = get_user_field(UJ_MITOSHEET_TELEMETRY) 
    return telemetry if telemetry is not None else False

def identify() -> None:
    """
    Helper function for identifying a user. We just take
    their python version, mito version, and email.
    """
    if not telemetry_turned_on():
        return

    static_user_id = get_user_field(UJ_STATIC_USER_ID)
    user_email = get_user_field(UJ_USER_EMAIL)
    feedbacks_v2 = get_user_field(UJ_FEEDBACKS_V2)

    params = {
        'version_mitoai': __version__,
        'email': user_email,
        UJ_FEEDBACKS_V2: feedbacks_v2
    }

    if not is_running_test():
        # TODO: If the user is in JupyterLite, we need to do some extra work.
        # You can see this in the mitosheet package. 
        try:
            analytics.identify(static_user_id, params)
        except Exception as e:
            pass


def log(
        log_event: str, 
        params: Optional[Dict[str, Any]]=None, 
        error: Optional[Exception]=None, 
    ) -> None:
    """
    This function is the entry point for all logging. 

    If telemetry is not turned off and we are not running tests,
    we log the ai event
    """

    final_params: Dict[str, Any] = params or {}
    
    # Then, make sure to add the user email
    final_params['email'] = get_user_field(UJ_USER_EMAIL)

    # Add the error if it exists
    if error is not None:
        final_params['error'] = str(error)

    # Finally, do the acutal logging. We do not log anything when tests are
    # running, or if telemetry is turned off
    if not is_running_test() and telemetry_turned_on():
        # TODO: If the user is in JupyterLite, we need to do some extra work.
        # You can see this in the mitosheet package. 
        try:
            analytics.track(
                get_user_field(UJ_STATIC_USER_ID), 
                log_event, 
                final_params
            )
        except Exception as e:
            pass
        

    # If we want to print the logs for debugging reasons, then we print them as well
    if PRINT_LOGS:
        print(
            log_event, 
            final_params
        )

    # TODO: Eventually we want to hook this up to the mito log uploader 
    # so enterprises can log usage if they want to.